#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
from __future__ import absolute_import

import glob
import os.path

import click.testing

import fortios_xutils.cli as TT
import fortios_xutils.firewall as F
import fortios_xutils.network as N
import fortios_xutils.parser as P
import tests.common as C


def abspaths(paths):
    return [os.path.abspath(p) for p in paths]


class Base(C.TestCase):

    sources = abspaths(C.list_res_files("show_configs", "*.txt"))
    cpaths = abspaths(C.list_res_files("parsed",
                                       os.path.join('*', P.ALL_FILENAME)))
    fpaths = abspaths(
        C.list_res_files("firewall",
                         os.path.join('*', F.FWP_TABLE_FILENAME))
    )
    npaths = abspaths(C.list_res_files("networks", "graph.yml"))

    # .. seealso:: tests/res/show_configs/*.txt
    ref_ips = ["192.168.122.10/24", "192.168.1.10/24"]
    ref_fas = set(("0.0.0.0/32",
                   "192.168.3.3/32",
                   "192.168.3.1/32",
                   "192.168.3.5/32",
                   "192.168.122.0/24",
                   "192.168.122.1/32",
                   "192.168.1.0/24",
                   "192.168.2.3/32",
                   "192.168.2.2/32",
                   "192.168.5.202/32",
                   "192.168.2.1/32",
                   "192.168.5.201/32",
                   "192.168.5.200/32"))


class TestCliCases(Base):

    def setUp(self):
        self.runner = click.testing.CliRunner()

    def test_10_parse__single_input(self):
        outdir = "out"
        for src in self.sources:
            with self.runner.isolated_filesystem():
                res = self.runner.invoke(TT.parse, [src, "-O", outdir])
                self.assertEqual(res.exit_code, 0)
                self.assertFalse(res.output)

                for fname in (P.METADATA_FILENAME, P.ALL_FILENAME):
                    files = glob.glob(os.path.join(outdir, '*', fname))
                    self.assertTrue(files)

    def test_12_parse__multi_inputs(self):
        outdir = "out"
        with self.runner.isolated_filesystem():
            res = self.runner.invoke(TT.parse, ["-O", outdir] + self.sources)
            self.assertEqual(res.exit_code, 0)
            self.assertFalse(res.output)

            for fname in (P.METADATA_FILENAME, P.ALL_FILENAME):
                files = glob.glob(os.path.join(outdir, '*', fname))
                self.assertTrue(files)

    def test_20_search__single_input(self):
        query = "configs[?config=='system interface'].edits[].ip"
        for cpath in self.cpaths:
            res = self.runner.invoke(TT.search, [cpath, "-P", query])
            self.assertEqual(res.exit_code, 0)
            self.assertTrue(res.output)

    def test_22_search__multi_inputs(self):
        query = "configs[?config=='system interface'].edits[].ip"
        res = self.runner.invoke(TT.search, ["-P", query] + self.cpaths)
        self.assertEqual(res.exit_code, 0)
        self.assertTrue(res.output)

    def test_30_network_collect__single_input(self):
        outdir = "out"
        for cpath in self.cpaths:
            with self.runner.isolated_filesystem():
                res = self.runner.invoke(TT.network_collect,
                                         ["-O", outdir, cpath])
                self.assertEqual(res.exit_code, 0)
                self.assertFalse(res.output)

                files = glob.glob(os.path.join(outdir, '*', N.NET_FILENAME))
                self.assertTrue(files)
                self.assertEqual(len(files), 1)

    def test_32_network_collect__multi_inputs(self):
        outdir = "out"
        with self.runner.isolated_filesystem():
            res = self.runner.invoke(TT.network_collect,
                                     ["-O", outdir] + self.cpaths)
            # self.assertEqual(res.exit_code, 0, res)
            self.assertFalse(res.output)

            files = glob.glob(os.path.join(outdir, '*', N.NET_FILENAME))
            self.assertTrue(files)
            self.assertEqual(len(files), 2)

    def test_40_network_compose__single_input(self):
        opath = os.path.join("out", N.NET_ALL_FILENAME)
        for cpath in self.cpaths:
            with self.runner.isolated_filesystem():
                res = self.runner.invoke(TT.network_compose,
                                         ["-o", opath, cpath])
                self.assertEqual(res.exit_code, 0)
                self.assertFalse(res.output)
                self.assertTrue(os.path.exists(opath))

    def test_42_network_compose__multi_inputs(self):
        opath = os.path.join("out", N.NET_ALL_FILENAME)
        with self.runner.isolated_filesystem():
            res = self.runner.invoke(TT.network_compose,
                                     ["-o", opath] + self.cpaths)
            self.assertEqual(res.exit_code, 0)
            self.assertFalse(res.output)
            self.assertTrue(os.path.exists(opath))

    def test_50_firewall_policy_save_and_search__single_input(self):
        outdir = "out"
        for cpath in self.cpaths:
            with self.runner.isolated_filesystem():
                res = self.runner.invoke(TT.firewall_policy_save,
                                         ["-O", outdir, cpath])
                self.assertEqual(res.exit_code, 0)
                self.assertFalse(res.output)

                files = glob.glob(os.path.join(outdir, '*',
                                               F.FWP_TABLE_FILENAME))
                self.assertTrue(files)
                self.assertEqual(len(files), 1)

                # not found
                ipa = "127.0.0.1"
                res = self.runner.invoke(TT.firewall_policy_search,
                                         ["-i", ipa, files[0]])
                self.assertEqual(res.exit_code, 0)
                self.assertTrue(res.output)
                self.assertEqual(res.output, "[]\n")

                # found
                ipa = "192.168.122.3"
                res = self.runner.invoke(TT.firewall_policy_search,
                                         ["-i", ipa, files[0]])
                self.assertEqual(res.exit_code, 0)
                self.assertTrue(res.output)
                self.assertNotEqual(res.output, "[]\n")

    def test_52_firewall_policy_save_and_search__multi_inputs(self):
        outdir = "out"
        with self.runner.isolated_filesystem():
            res = self.runner.invoke(TT.firewall_policy_save,
                                     ["-O", outdir] + self.cpaths)
            self.assertEqual(res.exit_code, 0)
            self.assertFalse(res.output)

            files = glob.glob(os.path.join(outdir, '*',
                                           F.FWP_TABLE_FILENAME))
            self.assertTrue(files)
            self.assertEqual(len(files), 2)

            # not found
            ipa = "127.0.0.1"
            for fdb in files:
                res = self.runner.invoke(TT.firewall_policy_search,
                                         ["-i", ipa, fdb])
                self.assertEqual(res.exit_code, 0)
                self.assertTrue(res.output)
                self.assertEqual(res.output, "[]\n")

            # found
            ipa = "192.168.122.3"
            for fdb in files:
                res = self.runner.invoke(TT.firewall_policy_search,
                                         ["-i", ipa, fdb])
                self.assertEqual(res.exit_code, 0)
                self.assertTrue(res.output)
                self.assertNotEqual(res.output, "[]\n")

    def test_70_network_find_paths__not_found(self):
        (src, dst) = ("127.0.0.1", "192.168.122.2")
        for npath in self.npaths:
            res = self.runner.invoke(TT.network_find_paths, [npath, src, dst])
            self.assertEqual(res.exit_code, 0)
            self.assertTrue(res.output)
            self.assertEqual(res.output, "[]\n")

    def test_72_network_find_paths__found(self):
        (src, dst) = ("192.168.122.2", "192.168.5.10")
        for npath in self.npaths:
            res = self.runner.invoke(TT.network_find_paths, [npath, src, dst])
            self.assertEqual(res.exit_code, 0)
            self.assertTrue(res.output)
            self.assertNotEqual(res.output, "[]\n")

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
"""Utility functions
"""
from __future__ import absolute_import

import collections.abc
import datetime
import glob
import hashlib
import os.path
import os

import anyconfig
import jmespath


_ENCODINGS = ("utf-8", "shift-jis")


def ensure_dir_exists(filepath, subdir=None):
    """Ensure dir for given file path `filepath` exists

    :param filepath: File path might be created later
    :param subdir: Sub dir
    :return: A dir path
    """
    tdir = os.path.dirname(filepath)
    if subdir:
        tdir = os.path.join(tdir, subdir)

    if not os.path.exists(tdir):
        os.makedirs(tdir)

    return tdir


def is_str(obj):
    """
    >>> is_str(u"foo")
    True
    >>> is_str("192.168.122.1/24")
    True
    >>> import ipaddress
    >>> is_str(ipaddress.ip_interface("192.168.122.1/24"))
    False
    """
    return isinstance(obj, (str, collections.abc.ByteString))


def timestamp(date_=False):
    """Generate timestamp string.

    :param date_: datetime.datetime object or False (default)
    """
    return (date_ or datetime.datetime.now()).strftime("%F_%H_%M_%S")


def checksum(filepath, hcnstrctr=hashlib.md5, enc="utf-8"):
    """Compute the checksum of given file, `filepath`

    :raises: OSError, IOError
    """
    return hcnstrctr(open(filepath).read().encode(enc)).hexdigest()


def get_subdir(filepath):
    """
    :param filepath: Path to the input file

    >>> get_subdir("/tmp/a/b/c/d.yml")
    'c'
    >>> get_subdir("/tmp/x.json")
    'tmp'
    >>> get_subdir("/")
    ''
    >>> get_subdir("a.yml")
    ''
    """
    return os.path.split(os.path.dirname(filepath))[-1]


def get_io_path(filepath, filename, outdir=False):
    """
    :param filepath: A paths to input file
    :param filename: Base file name of output
    :param outdir: Dir to save outputs, same dir input files exist by default.

    :return: A pair of paths of input and output files

    >>> get_io_path("/a/b/c/d.json", "x.yml")
    ('/a/b/c/d.json', '/a/b/c/x.yml')
    >>> get_io_path("/a/b/c/d.json", "x.yml", "/0/1/2")
    ('/a/b/c/d.json', '/0/1/2/c/x.yml')
    """
    if outdir:
        outdir = os.path.join(outdir, get_subdir(filepath))
    else:
        outdir = os.path.dirname(filepath)

    return (filepath, os.path.join(outdir, filename))


def get_io_paths(filepaths, filename, outdir=False):
    """
    :param filepaths:
        A list of paths to the JSON files contains the parsed results of
        fortigate's 'show configuration' outputs

    :param filename: Base file name of output
    :param outdir: Dir to save outputs [same dir input files exist]

    :yield: A pair of input and output paths
    """
    for fpath in filepaths:
        # Compute `outdir` for each `fpath` to save results separately.
        yield get_io_path(fpath, filename, outdir=outdir)


def try_ac_load(filepath, type_=None, encodings=_ENCODINGS):
    """
    Try to open and load `filepath` using anyconfig.load.

    :param filepath: File path to load
    :param type_: File type of the file, `filepath`
    :param encodings: Character set encodings to try

    :return:
        A mapping object or sequence objects loaded from `filepath` or None
    """
    for enc in encodings:
        try:
            with open(filepath, encoding=enc) as inp:
                return anyconfig.load(inp, ac_parser=type_)
        except UnicodeDecodeError:
            pass

    return None


def save_file(data, filepath, **ac_opts):
    """
    An wrapper for anyconfig.dump.

    :param data: Data to save
    :param filepath:  Path to output file
    :param ac_opts: Keyword arguments will be given to anyconfig.dump
    """
    ensure_dir_exists(filepath)
    anyconfig.dump(data, filepath, **ac_opts)


def search(jmespath_exp, data):
    """
    Just an wrapper for :func:`jmespath.search`

    >>> search("[].a[]",
    ...        [dict(a=[1, 2], b=2), dict(b=3, c=4), dict(a=[3,4])])
    [1, 2, 3, 4]
    >>> search("[].x", [dict(a=[1, 2], b=2)])
    []
    """
    return jmespath.search(jmespath_exp, data)


def expand_glob_paths_itr(filepaths, marker='*'):
    """
    :param filepaths: A list of file paths
    """
    for fpath in filepaths:
        if marker in fpath:
            for path in sorted(glob.glob(fpath)):
                yield path
        else:
            yield fpath

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2020 Satoru SATOH <satoru.satoh@gmail.com>.
# SPDX-License-Identifier: MIT
#
r"""Find network nodes.
"""
from __future__ import absolute_import

from . import firewall, network, parser, utils

# pylint: disable=unused-import
from .network import NODE_TYPES, NODE_ANY  # noqa: F401


def parse_and_save_show_configs(filepaths, outdir):
    """
    :param filepaths:
        An iterable yields path might contains '*' (glob) pattern will be
        expanded to a list of paths of files each gives a str or
        :class:`pathlib.Path` object represents file path contains 'show
        full-configuration` or any other 'show ...' outputs
    :param outdir:
        Dir to save parsed results as JSON files [out/ relative to filepaths]

    :return:
        A list of a tuple of (input file path, mapping object contains parsed
        results)
    :raises: IOError, OSError
    """
    fsit = utils.expand_glob_paths_itr(filepaths)
    return list(parser.parse_show_configs_and_dump_itr(fsit, outdir))


def _query_json_files_itr(filepaths, path_exp):
    """
    :param filepaths:
        An iterable yields a path might contain '*' (glob) pattern will be
        expanded to a list of paths to JSON input files to query using JMESPath
        expression.
    :param path_exp: JMESPath expression to query

    :yields: A tuple of (input file path, a mapping object gives the result)
    """
    for filepath in utils.expand_glob_paths_itr(filepaths):
        cnf = parser.load(filepath)
        res = parser.jmespath_search(path_exp, cnf,
                                     has_vdoms_=parser.has_vdom(cnf))
        yield (filepath, res)


def query_json_files(filepaths, path_exp):
    """
    :param filepaths:
        An iterable yields a path might contain '*' (glob) pattern will be
        expanded to a list of paths to JSON input files to query using JMESPath
        expression.
    :param path_exp: JMESPath expression to query

    :return: A tuple of (input file path, a mapping object gives the result)
    """
    return [dict(filepath=t[0], results=t[1]) for t
            in _query_json_files_itr(filepaths, path_exp)]


def collect_networks(filepaths, prefix=network.NET_MAX_PREFIX):
    """
    :param filepaths:
        An iterable yields a path might contain '*' (glob) pattern will be
        expanded to a list of paths to JSON input files which contains parsed
        and structured results of fortigate's configuration outputs
    :param prefix: Max network prefix to collect

    :return: A tuple of (input file path, a mapping object gives the result)
    """
    fsit = utils.expand_glob_paths_itr(filepaths)
    return network.collect_networks_from_config_files(fsit, prefix=prefix)


def collect_and_save_networks(filepaths, outdir=False,
                              prefix=network.NET_MAX_PREFIX):
    """
    :param filepaths:
        An iterable yields a path might contain '*' (glob) pattern will be
        expanded to a list of paths to JSON input files which contains parsed
        and structured results of fortigate's configuration outputs
    :param outdir: Dir to save outputs [same dir input files exist]
    :param prefix: Max network prefix to collect

    :return: A tuple of (input file path, a mapping object gives the result)
    """
    fsit = utils.expand_glob_paths_itr(filepaths)
    opts = dict(outdir=outdir, prefix=prefix)

    return network.collect_and_save_networks_from_config_files(fsit, **opts)


def compose_networks(filepaths):
    """
    Compose a network graphs consist of nodes and edges (node and network
    links) information collected from fortigate's parsed configuration file.

    :param filepaths:
        A list of path to the JSON file contains network graph data

    :return: A graph data contains metadata, nodes and links data
    """
    fsit = utils.expand_glob_paths_itr(filepaths)
    return network.compose_network_files(fsit)


def compose_and_save_networks(filepaths, outpath=False):
    """
    Compose a network graphs consist of nodes and edges (node and network
    links) information collected from fortigate's parsed configuration file.

    :param filepaths:
        A list of path to the JSON file contains network graph data
    :param outpath: Output file path

    :return: A graph data contains metadata, nodes and links data
    """
    fsit = utils.expand_glob_paths_itr(filepaths)
    return network.compose_and_save_network_files(fsit, outpath=outpath)


def make_firewall_policy_table(filepath):
    """
    :param filepath:
        A path might to JSON input files which contains parsed and structured
        results of fortigate's configuration outputs

    :return: A :class:`pandas.DataFrame` object
    """
    return firewall.make_firewall_policy_table(filepath)


def make_firewall_policy_tables(filepaths):
    """
    :param filepaths:
        An iterable yields a path might contain '*' (glob) pattern will be
        expanded to a list of paths to JSON input files which contains parsed
        and structured results of fortigate's configuration outputs

    :return: A list of :class:`pandas.DataFrame` object
    """
    return firewall.make_firewall_policy_tables(filepaths)


def make_and_save_firewall_policy_tables(filepaths, outdir=False):
    """
    :param filepaths:
        An iterable yields a path might contain '*' (glob) pattern will be
        expanded to a list of paths to JSON input files which contains parsed
        and structured results of fortigate's configuration outputs

    :return: A list of :class:`pandas.DataFrame` object
    """
    return firewall.make_and_save_firewall_policy_tables(
        filepaths, outdir=outdir
    )


def load_firewall_policy_table(filepath):
    """
    :param filepath: Path to the JSON file contains fortigate's configurations

    :return: A :class:`pandas.DataFrame` object
    """
    return firewall.load_firewall_policy_table(filepath)


def search_firewall_policy_table_by_addr(ip_s, tbl_df):
    """
    :param ip_s: A str represents an IP address
    :param tbl_df:
        A :class:`pandas.DataFrame` object contains ip addresses in the columns
        have one or some of `addrs_cols`.

    :return: A list of mappping objects contains results
    """
    return firewall.search_by_addr_1(ip_s, tbl_df)

# vim:sw=4:ts=4:et:
